/*
 * client.c: A very, very primitive HTTP client.
 * 
 * To run, try: 
 *      client www.cs.wisc.edu 80 /
 *
 * Sends one HTTP request to the specified HTTP server.
 * Prints out the HTTP response.
 *
 * CS537: For testing your server, you will want to modify this client.  
 * For example:
 * 
 * You may want to make this multi-threaded so that you can 
 * send many requests simultaneously to the server.
 *
 * You may also want to be able to request different URIs; 
 * you may want to get more URIs from the command line 
 * or read the list from a file. 
 *
 * When we test your server, we will be using modifications to this client.
 *
 */

#include "cs537.h"
#include "mywrappers.h"


int port;
char *host;

/*
 * Send an HTTP request for the specified file 
 */
void clientSend(int fd, char *filename)
{
  char buf[MAXLINE];
  char hostname[MAXLINE];

  Gethostname(hostname, MAXLINE);

  /* Form and send the HTTP request */
  sprintf(buf, "GET %s HTTP/1.1\n", filename);
  sprintf(buf, "%shost: %s\n\r\n", buf, hostname);
  Rio_writen(fd, buf, strlen(buf));
}
  
/*
 * Read the HTTP response and print it out
 */
void clientPrint(int fd, char *filename)
{
  rio_t rio;
  char buf[MAXBUF];  
  int length = 0;
  int n;
  
  Rio_readinitb(&rio, fd);

  /* Read and display the HTTP Header */
  n = Rio_readlineb(&rio, buf, MAXBUF);
  while (strcmp(buf, "\r\n") && (n > 0)) {
    printf("Header: %s", buf);
    n = Rio_readlineb(&rio, buf, MAXBUF);

    /* If you want to look for certain HTTP tags... */
    if (sscanf(buf, "Content-Length: %d ", &length) == 1) {
      printf("Length = %d\n", length);
    }
  }

  /* Read and display the HTTP Body */
	if(strstr(filename, "cgi") != NULL)
	{	n = Rio_readlineb(&rio, buf, MAXBUF);
		while (n > 0) {
			printf("%s", buf);
			n = Rio_readlineb(&rio, buf, MAXBUF);
		}
	}
}

void threadHandle(void * filename);
int main(int argc, char *argv[])
{	int i;
	int threadc;
		
	if (argc < 4) {
		fprintf(stderr, "Usage: %s <host> <port> <filename>\n", argv[0]);
		exit(1);
	}

	host = argv[1];
	port = atoi(argv[2]);
	threadc = argc - 3;
	
	char *filename[threadc];
	//printf("threadc %d", threadc); fflush(stdout);
	for(i = 0; i < threadc; i++)
	{	filename[i] = argv[i + 3];
	}

	pthread_t threadv[threadc];
	
	/* Open a single connection to the specified host and port */
  	for(i = 0; i < threadc; i++)
	{	pthread_create(&threadv[i], NULL, (void *)&threadHandle, (void *)filename[i]);
	}

	for(i = 0; i < threadc; i++)
	{	pthread_join(threadv[i], NULL);
	}

	exit(0);
}

void threadHandle(void * filename)
{	int clientfd;
	
	clientfd = Open_clientfd(host, port);
  
  	clientSend(clientfd, (char *)filename);
	clientPrint(clientfd, (char *)filename);
    
  	Close(clientfd);
}
