#include "mywrappers.h"

/*********************************************
 * Wrappers for time functions
 ********************************************/
void Gettimeofday(timeval_t *tv, timezone_t *tz)
{	if(gettimeofday(tv, tz) == -1)
	{	perror("timeofdayerror: ");
		exit(1);
	}
}

/*********************************************
 * Wrappers for memory allocation functions
 ********************************************/

/* Malloc wrapper */
void *Malloc(size_t size)
{	void *ptr;
	if((ptr = malloc(size)) != NULL)
	{	return ptr;
	}
	else
	{	fprintf(stderr, "malloc: memory request failed");
		exit(1);
	}	
}


int retval;
/*********************************************
 * Wrappers for POSIX thread control functions
 ********************************************/

/* Pthread_create wrapper */
void Pthread_create(pthread_t *thread, pthread_attr_t *attr,
             			 void *(*start_routine)(void*), void *arg)
{	if(pthread_create(thread, attr, start_routine, arg) != 0)
	{	fprintf(stderr, "pthread_create: failed to create");
		exit(1);
	}
}


/* Pthread_detach wrapper */
void Pthread_detach(pthread_t thread)
{	if(pthread_detach(thread) != 0)
	{	fprintf(stderr, "pthread_detach: failed to detach");
		exit(1);
	}
}

/* Pthread_mutex_lock wrapper */
void Pthread_mutex_lock(pthread_mutex_t  *mutex)
{	if((retval = pthread_mutex_lock(mutex)) != 0)
	{	switch(retval)
		{	case EINVAL:
				fprintf(stderr, "mutex_lock: mutex not been properly initialized");
				break;
			case EDEADLK:
				fprintf(stderr, "mutex_lock: mutex  is  already  locked  by calling  thread");
				break;
			default:
				fprintf(stderr, "mutex_lock: unknown error");
		}
		exit(1);
	}
}

/* Pthread_mutex_unlock wrapper */
void Pthread_mutex_unlock(pthread_mutex_t  *mutex)
{	if((retval = pthread_mutex_unlock(mutex)) != 0)
	{	switch(retval)
		{	case EINVAL:
				fprintf(stderr, "mutex_unlock: mutex not been properly initialized");
				break;
			case EPERM:
				fprintf(stderr, "mutex_unlock: calling thread does not own mutex");
				break;
			default:
				fprintf(stderr, "mutex_unlock: unknown error");
		}
		exit(1);
	}
}

/* Pthread_cond_wait wrapper */
void Pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex)
{	if(pthread_cond_wait(cond, mutex) != 0)
	{	fprintf(stderr, "pthread_cond_wait: no fucking clue");
		exit(1);
	}
}

/* Pthread_cond_signal wrapper */
void Pthread_cond_signal(pthread_cond_t *cond)
{	if(pthread_cond_signal(cond) != 0)
	{	fprintf(stderr, "pthread_cond_signal: no fucking clue");
		exit(1);
	}
}

/*********************************
 *  INITIALIZER WRAPPERS (IF NEEDED)
 *********************************/

/* Pthread_mutex_init wrapper */
void Pthread_mutex_init(pthread_mutex_t  *mutex,  const pthread_mutexattr_t *mutexattr)
{	if(pthread_mutex_init(mutex, mutexattr) != 0)
	{	fprintf(stderr, "pthread_mutex_init: failed to init");
		exit(1);
	}
}

/* Pthread_cond_init wrapper */
void Pthread_cond_init(pthread_cond_t *cond, pthread_condattr_t *cond_attr)
{	if(pthread_cond_init(cond, cond_attr) != 0)
	{	fprintf(stderr, "pthread_cond_init: failed to init");
		exit(1);
	}
}
