#include "cs537.h"
#include "request.h"
#include "mywrappers.h"
#include "my_thread.c"
#include <sys/time.h>

/* 
 * server.c: A very, very simple web server
 *
 * To run:
 *  server <portnum (above 2000)>
 *
 * Repeatedly handles HTTP requests sent to this port number.
 * Most of the work is done within routines written in request.c
 *
 * CS537 Students: You will be modifying this module
 */

timeval_t START_TIME;
int TOTAL_REQUESTS_DISPATCHED = 0;

enum {FCFS, SFF};

int *schedalgptr;

//thread variables
sthread_t *threadvptr;

//buffer variables
request_t **bufferptr;
int size  = 0;
int in    = 0;
int out   = 0;
int count = 0;

//synconization variables
pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t empty = PTHREAD_COND_INITIALIZER;
pthread_cond_t full  = PTHREAD_COND_INITIALIZER;

/*
 * CS537: Parse the new arguments too
 */

void setNext(const request_t *req, int schedalg);
void getNext(request_t *req);
void *threadHandle(void *sthread);
void sortBufferByFileSize();
void swapBuffers(int i, int j);
void requestcpy(request_t *dest, const request_t *src);

void getargs(int argc, char *argv[], int *port, int *threadc, int *buffc, int *schedalg)
{
	if (argc != 5) {
		fprintf(stderr, "Usage: %s <port> <threads> <buffers> <schedalg>\n", argv[0]);
		exit(1);
	}
	*port = atoi(argv[1]);
	if((*threadc = atoi(argv[2])) < 1)
	{	fprintf(stderr, "error: # of threads must be a postive integer");
		exit(1);
	}
	if((*buffc = atoi(argv[3])) < 1)
	{	fprintf(stderr, "error: # of buffers must be a postive integer");
		exit(1);
	}
	
	if(strcmp(argv[4], "FCFS") == 0)
	{	*schedalg = FCFS;}
	else if (strcmp(argv[4], "SFF") == 0)
	{	*schedalg = SFF; }
	else
	{	fprintf(stderr, "error: unrecognized schedule algorithm: %s", argv[4]);
	}
}


int main(int argc, char *argv[])
{	int i;
	int listenfd, clientlen;
	int port, threadc, buffc, schedalg;
	struct sockaddr_in clientaddr;

	getargs(argc, argv, &port, &threadc, &buffc, &schedalg);
	schedalgptr = &schedalg;
	
	/* Create buffer */
	request_t *buffer[buffc];
	bufferptr = buffer;
	size = buffc;

	/* CS537: Create some threads...*/
	sthread_t threadv[threadc];
	threadvptr = threadv;
	
	for(i = 0; i < threadc; i++)
	{	threadv[i].thread_id = i;
		threadv[i].thread_count = 0;
		threadv[i].static_count = 0;
		threadv[i].dynamic_count = 0;
	}	

	for(i = 0; i < threadc; i++)
	{	Pthread_create(&threadv[i].thread, NULL, (void *)&threadHandle, (void *)&threadv[i]);
	}
	
	request_t req;
	listenfd = Open_listenfd(port);
	Gettimeofday(&START_TIME, NULL);
	while (1) {
		clientlen = sizeof(clientaddr);
		req.connfd = Accept(listenfd, (SA *)&clientaddr, &clientlen);
	
		/*request started*/
		Gettimeofday(&req.arrival_time, NULL); 							//get arrival time
		timeval_t t = timesub(req.arrival_time, START_TIME);
		timecpy(&req.arrival_time, &t);
		req.trd_at_arrival = TOTAL_REQUESTS_DISPATCHED;					//get info for age

		if(schedalg == SFF)
		{	req.size = 0;
			requestFilesize(req.connfd, &req);
			if(req.size != -1) setNext(&req, schedalg);
		} else {
			setNext(&req, schedalg);
		}
	}
}

/* setNext - producer
 * adds bc to next available cell in the buffer
 */
void setNext(const request_t *req, int schedalg)
{	Pthread_mutex_lock(&lock);
	while(count == size)	Pthread_cond_wait(&empty, &lock);
	
	/*put request in buffer*/
	bufferptr[in] = (request_t *)Malloc(sizeof(request_t));
	requestcpy(bufferptr[in], req);
	in = (in + 1) % size;
	count++;
	if(schedalg == SFF) sortBufferByFileSize();
	Pthread_cond_signal(&full);
	Pthread_mutex_unlock(&lock);
}

/* getNext - consumer
 * get the next available request from buffer
 */
 void getNext(request_t *req)
{	timeval_t tv;
	Pthread_mutex_lock(&lock);
	while(count == 0)	Pthread_cond_wait(&full, &lock);
	
	/*get request from buffer*/	
	Gettimeofday(&tv, NULL);											
	requestcpy(req, bufferptr[out]);
	req->age = TOTAL_REQUESTS_DISPATCHED++ - req->trd_at_arrival;							
	printf("age: %d TRD: %d trdaa: %d", req->age, TOTAL_REQUESTS_DISPATCHED, req->trd_at_arrival);

	tv = timesub(timesub(tv, req->arrival_time), START_TIME);		//get dispatch time
	timecpy(&req->dispatch_time, &tv);
	
	free(bufferptr[out]);
	bufferptr[out] = NULL;

	out = (out + 1) % size;
	count--;
	Pthread_cond_signal(&empty);
	Pthread_mutex_unlock(&lock);
}

void *threadHandle(void *sthread)
{	request_t req;
	sthread_t *thr = (sthread_t *)sthread;
	while(1)
	{	getNext(&req);
		thr->thread_count++;		/*thread num processes ++*/;
		printf("thread %d(%s %d) start\n", thr->thread_id, req.filename, req.size); fflush(stdout);
		if(*schedalgptr == FCFS) 
		{	requestHandle(req.connfd, thr, &req);
		} else {
			requestHandleSFF(req.connfd, thr, &req);
		}
		printf("thread %d complete\n", thr->thread_id); fflush(stdout);
		
		Close(req.connfd);
	}
}

void sortBufferByFileSize()
{	int i, j;
	for(i = 0; i < count - 1; i++)
	{	for(j = i; j < count - 1 - i; j++)
		{	int a = (j + out) % size;
			int b = (j + out + 1) % size;
			if(bufferptr[a]->size > bufferptr[b]->size)
			{	swapBuffers(a, b);
			}
		}
	}
}

void swapBuffers(int i, int j)
{	request_t *temp;
	temp = bufferptr[i];
	bufferptr[i] = bufferptr[j];
	bufferptr[j] = temp;
}

void requestcpy(request_t *dest, const request_t *src)
{	dest->connfd			= src->connfd;
	dest->trd_at_arrival	= src->trd_at_arrival;
	dest->age				= src->age;
	dest->size			= src->size;
	
	timecpy(&dest->arrival_time, &src->arrival_time);
	timecpy(&dest->dispatch_time, &src->dispatch_time);
	timecpy(&dest->read_time, &src->read_time);
	timecpy(&dest->completion_time, &src->completion_time);
	
	dest->is_static		= src->is_static;
	strcpy(dest->filename, src->filename);
	strcpy(dest->cgiargs, src->cgiargs);
}
